﻿using System.Runtime.InteropServices;

namespace UEM_COM
{
    /// <summary>
    /// Класс-обертка, для вызовов нативных функций Clscrfl_SAM.dll
    /// </summary>
    internal sealed class Clscrfl
    {
        /// <summary>
        /// Получение Singleton объекта экземпляра класса Clscrfl 
        /// </summary>
        internal static readonly Clscrfl Instance = new Clscrfl();

        private Clscrfl()
        {
        }

        unsafe private void* pReader = (void*)0;

        #region Прототипы для нативных функций из Clscrfl_SAM.dll

        [DllImport("Clscrfl_SAM.dll")]
        unsafe private static extern int CLSCRF_Create(void** pReader);

        [DllImport("Clscrfl_SAM.dll")]
        unsafe private static extern int CLSCRF_Destroy(void** pReader);

        [DllImport("Clscrfl_SAM.dll")]
        unsafe private static extern int CLSCRF_OpenRS(void* pReader,
                                                uint dwPortNumber,
                                                uint dwBaudrate,
                                                uint dwLogFile );
        
        [DllImport("Clscrfl_SAM.dll")]
        unsafe private static extern int CLSCRF_GetState(void* pReader,
                                                out int state);

        [DllImport("Clscrfl_SAM.dll")]
        unsafe private static extern int CLSCRF_Sound(void* pReader,
                                                byte nBeepCount);

        [DllImport("Clscrfl_SAM.dll")]
        unsafe private static extern int CLSCRF_Led(void* pReader,
                                                byte bBlinkColor,
                                                byte bBlinkCount,
                                                byte bPostColor);

        [DllImport("Clscrfl_SAM.dll")]
        unsafe private static extern int CLSCRF_Get_Mfrc_Version(void* pReader,
                                                System.IntPtr pbVersion);

        #endregion Прототипы для нативных функций из Clscrfl_SAM.dll

        #region Вызовы оберток для нативных функций из Clscrfl_SAM.dll

        /// <summary>
        /// Создать интерфейс.
        /// Функция должна вызываться 1 раз в начале работы приложения.
        /// </summary>
        /// <returns>Возвращаемое значение:  0   - успешное выполнение
        /// иначе - ошибка при выполнении</returns>
        internal int Create()
        {
            int status = -1;

            unsafe
            {
                fixed (void** ppReader = &pReader)
                {
                    status = CLSCRF_Create(ppReader);
                }
            }

            return status;
        }

        /// <summary>
        /// Уничтожить интерфейс.
        /// Функция должна вызываться 1 раз в конце работы приложения.
        /// </summary>
        /// <returns>Возвращаемое значение:  0   - успешное выполнение
        /// иначе - ошибка при выполнении</returns>
        internal int Destroy()
        {
            int status = -1;

            unsafe
            {
                fixed (void** ppInface = &pReader)
                {
                    status = CLSCRF_Destroy(ppInface);
                }
            }

            return status;
        }

        /// <summary>
        /// Открыть порт.
        /// </summary>
        /// <param name="dwPortNumber">Номер порта</param>
        /// <param name="dwBaudrate">Скорость шины RS485</param>
        /// <param name="dwLogFile">Создание лога</param>
        /// <returns>Возвращаемое значение:  0   - успешное выполнение
        /// иначе - ошибка при выполнении</returns>
        internal int OpenCOM(uint dwPortNumber, uint dwBaudrate, uint dwLogFile)
        {
            int status = -1;

            unsafe
            {
                status = CLSCRF_OpenRS(pReader, dwPortNumber-1, dwBaudrate, dwLogFile);
            }

            return status;
        }

        /// <summary>
        /// Произвести звуковую индикацию.
        /// </summary>
        /// <param name="nBeepCount">Количество сигналов</param>
        /// <returns>Возвращаемое значение:  0   - успешное выполнение
        /// иначе - ошибка при выполнении</returns>
        internal int Sound(byte nBeepCount)
        {
            int status = -1;

            unsafe
                {
                    status = CLSCRF_Sound(pReader, nBeepCount);
                }

            return status;
        }

        /// <summary>
        /// Произвести индикацию светодиодами.
        /// </summary>
        /// <param name="bBlinkColor">Цвет сигнала</param>
        /// <param name="bBlinkCount">Количество сигналов</param>
        /// <param name="bPostColor">Цвет после сигнала</param>
        /// <returns>Возвращаемое значение:  0   - успешное выполнение
        /// иначе - ошибка при выполнении</returns>
        internal int Led(byte bBlinkColor, byte bBlinkCount, byte bPostColor)
        {
            int status = -1;

                unsafe
                {
                    status = CLSCRF_Led(pReader, bBlinkColor, bBlinkCount, bPostColor);
                }

            return status;
        }

        /// <summary>
        /// Получить версию микропрограммы
        /// </summary>
        /// <param name="pbVersion">Указатель на массив из 6 байтов</param>
        /// <returns>Возвращаемое значение:  0   - успешное выполнение
        /// иначе - ошибка при выполнении</returns>
        internal int GetVersion(ref byte[] pbVersion)
        {
            int status = -1;

            var pbVer = Marshal.AllocHGlobal(6);
            unsafe
            {
                status = CLSCRF_Get_Mfrc_Version(pReader, pbVer);
            }
            if (status == 0)
                Marshal.Copy(pbVer, pbVersion, 0, 6);
            Marshal.FreeHGlobal(pbVer);

            return status;
        }

        #endregion  Вызовы оберток для нативных функций из Clscrfl_SAM.dll
    }
}
